%%
% Matlab script that explains how to solve (SP) in cvx
%
% Graph type: flag_p = 1; % Erdos-Renyi network
%             flag_p = 2; % path network
%             flag_p = 3; % disconnected randomly generated network (with N = 50 nodes)
%

clear, clc, close all

% number of nodes
N = 50;

flag_p = 1; % for Erdos-Renyi network
% flag_p = 2; % for path network
% flag_p = 3; % for disconnected randomly generated network (with N = 50 nodes)

if flag_p == 1,
    
    % Edge probability for Erdos-Renyi network
    p = 1.05*log(N)/N;
    
    % Graph Laplacian of an Erdos-Renyi plant network
    Lp = random_graph_generator(N,p);
    eigLp = sort(eig(Lp),'ascend');

    % Check if a plant graph is connected 
    while abs(eigLp(2)) <= 1.e-12

        Lp = random_graph_generator(N,p);
        eigLp = sort(eig(Lp),'ascend');

        if abs(eigLp(2)) >= 1.e-12
            
            break;
            
        end
        
    end
     
elseif flag_p == 2

    % Construct incidence matrix of a path graph
    Ep = toeplitz([1 -1 zeros(1,N-2)],[1 zeros(1,N-2)]);
    
    % plant graph Laplacian
    Lp = Ep*Ep';
      
elseif flag_p == 3,
    
    N = 50;
    Length = 10;

    load pos;
    pos = pos(1:N,:);
    r = Length/5;
    reweighted_Max_Iter = 1;

    m = N*(N-1)/2;

    % determine the neighbors of each node
    idx = zeros(2,m);
    Lp = zeros(N,N); 
    stp = 0;

    % construct Lp
    for i = 1:N,

        for j = i+1:N,
            % if the distance between two nodes is less than r
            % then they are neighbors.

            if norm( pos(i,:) - pos(j,:), 2 ) <= r,
                
                stp = stp + 1;
                idx(:,stp) = [i j]';
                Lp(i,j) = -1;
                Lp(j,i) = -1;
                
            end

        end

    end

    for p = 1:N

       Lp(p,p) = -sum(Lp(p,:));

    end
    
else 
    
     error('Incorrect graph type')
     
end

% make sure Lp is sparse
Lp = sparse(Lp);

% Construct incidence matrix of the controller graph
% No overlap between plan and controller
% Ecomplete = [Ep, Ec]
d = diag(Lp);
S = N^2 - N + nnz(d) - nnz(Lp);
index = zeros(S/2,2);

c = 1;

for i = 1:N,
    
    for j = i+1:N,
        
        if Lp(i,j)==0,
            
            index(c,:) = [i,j];
            
            c = c + 1;
            
        end
        
    end
    
end

% incidence matrix of the controller graph
Ec = incmat(index',N);

% number of edges in the controller graph
m = size(Ec,2);
    
% state and control weights in LQR objective
% state weight - deviation from average
Q = eye(N) - (1/N)*ones(N,N);
% control weight
R = speye(N);

% "strengthened" state weight
Qp = Q + (1/N)*ones(N,N) + Lp*R*Lp; 

% compute Gp
Gp = Lp + (1/N)*ones(N,N);

% form the vector that appears in the objective function
qvec = sum(Ec.*(R*Ec))';

% Qphalf
Qphalf = sqrtm(Qp);
Qphalf = 0.5*real(Qphalf + Qphalf');

% choose gamma
if (flag_p == 1 || flag_p == 2) 
    
    % compute invGp
    invGp = Gp\eye(N);
    Yp = invGp*Q*invGp;
    
    % compute gamma_max
    gamma_max = max(abs(sum(Ec.*(Yp*Ec))));
    % regularization parameter gamma
    gamma = 0.5*full(gamma_max);

else
    
    % for random disconnected network
    gamma = 10;
    
end

% run cvx
tic
cvx_begin sdp 
    
    variable X(N,N) symmetric
    variable xcvx(m,1)
    
    minimize (trace(X) + qvec'*xcvx + gamma*sum(abs(xcvx)))
    subject to 
    [X, Qphalf; Qphalf, Gp + Ec*diag(xcvx)*Ec'] >= 0;
    
cvx_end
tcvx = toc;

% performance of sparsity-promoting controller
% obtained using cvx
Lcvx = Ec*diag(xcvx)*Ec';
Gcvx = Gp + Lcvx;
Jcvx = trace( Gcvx\(Q + Lcvx*R*Lcvx) );

% Visualize results
figure
% plot optimal vector of the edge weights (before polishing)
plot(1:m,xcvx,'*')
xlab = xlabel('edge index','interpreter','tex');
set(xlab,'FontName','cmmr10','FontSize',26)
ylab = ylabel('edge weights','interpreter','tex');
set(ylab,'FontName','cmmr10','FontSize',26)
h = get(gcf,'CurrentAxes');
set(h,'FontName','cmr10','FontSize',20,'xscale','lin','yscale','lin')
axis([0 m+1 0 1.1*max(xcvx)])
