% Sparsity-Promoting Optimal H_2 Controller for Undirected Networks
%
% Written by: Sepideh Hassan-Moghaddam and Mihailo R. Jovanovic
%
% May 2015
%
% Customized Algorithm based on Infeasible Primal-Dual Interior-Point Method
%
% Network Dynamics: qdot = -(Lp + Lx)*q + d
% 
% q - state of the network 
% d - white stochastic disturbance 
%
% Lp - graph Laplacian of the plant
% Lx = Ec*diag(x)*Ec' - graph Laplacian of the controller
% Ec - incidence matrix of the controller graph
% x  - vector of the edge weights
%
% Optimization Problem
%
% minimize <Qp,(Gp + Lx)^{-1}> + <R,Lx> + gamma*sum(x_i)
%
% subject to Gp + Lx > 0
% 
% Syntax:
%
% out = graphsp_IP(Lp,Ec,Q,R,gamma,opts)
%
% Inputs: (1) plant graph Laplacian Lp
%             incidence matrix of the controller graph Ec
%             state weight Q
%             controller weight R
%             sparsity-promoting parameter gamma
%
%         (2) options
%
%             opts.eps_dg   - tolerance on duality gap
%             opts.eps_pres  - tolerance on primal residual
%             opts.maxit    - maximum number of IP iterations
%             opts.maxit_pcg - maximum number of PCG iterations
%             opts.delta_pcg_aff  - parameter in pcg algorithm
%             opts.delta_pcg - parameter pcg algorithm
%             opts.method   - method used for computing search direction
%
%             Search Directions Obtained Using: 
%
%             Cholesky Factorization: opts.method = 'chol'
%             PCG with storing A; diagonal preconditioner: opts.method = 'pcg1'
%             PCG without storing A; diagonal preconditioner: opts.method = 'pcg2'
%
%             If opts argument is omitted, the default values are set to
%
%             opts.eps_dg   = 1.e-6
%             opts.eps_pres  = 1.e-6
%             opts.maxit    = 100
%             opts.maxit_pcg = 2000
%             opts.delta_pcg_aff  = 0.3
%             opts.delta_pcg = 0.3
%             opts.method   = 'pcg2' 
%
% Output:  out - output structure containing
%
%          out.x    - optimal vector of edge weights 
%          out.zp   - optimal z+
%          out.zm   - optimal z-
%          out.xp   - optimal x+
%          out.xm   - optimal x-
%          out.dg   - duality gap
%          out.dres - dual residual
%          out.pres - primal residual
%          out.s    - number of IP iterations
%          out.ind1 - number of PCG iterations for affine search direction
%          out.ind2 - number of PCG iterations for Mehrotra search direction
%
% Additional information:
%
% http://www.umn.edu/~mihailo/software/graphsp/


function out = graphsp_IP(Lp,Ec,Q,R,gamma,opts) 

% Initialization
if nargin < 5
    error('At least six input arguments are required.')
elseif nargin == 5
    opts = struct('eps_dg',1.e-6,'eps_pres',1.e-6,'maxit',100,'maxit_pcg',2000,'method','pcg2','delta_pcg_aff',0.3,'delta_pcg',0.3);
elseif nargin > 6
    error('Too many input arguments.')
end

% number of nodes
N = size(Ec,1);

% number of edges in the controller graph
% determines the number of optimization variables
m = size(Ec,2);

% Data preprocessing
eps_dg  = opts.eps_dg;  % tolerance on duality gap
eps_pres = opts.eps_pres; % tolerance on primal residual
maxit = opts.maxit;
maxit_pcg = opts.maxit_pcg;
flag = opts.method;
delta_pcg_aff = opts.delta_pcg_aff;
delta_pcg = opts.delta_pcg;

% sparse identity matrix
I = speye(N); 

% delta in IP method
delta = 3;

% tau for step size selection
if gamma <= 1.e-12
    
    % parameters for optimal centralized controller
    tau = 0.9; % parameter in step-size selection
    eps_dres = 0.1; % tolerance on dual residual
    flag = 'chol';
    
else
    
    % parameters for sparsity-promoting problem
    tau = 1; % parameter in step-size selection
    eps_dres = 1.e-3; % tolerance on dual residual
    
end
 
% "strengthened" plant graph Laplacian
Gp = Lp + (1/N)*ones(N,N);

% strengthened state weight 
Qp = Q + (1/N)*ones(N,N) + Lp*R*Lp; 

% form the vector that appears in the objective function
qvec = sum(Ec.*(R*Ec))';

% initialize feedback gain
xhat = 0.2*ones(m,1); 
x = xhat;
 
% initialize xp
xphat = 0.6*ones(m,1); 
xp = xphat;

% initialize xm
xmhat = 0.5*ones(m,1); 
xm = xmhat;

% compute invG
G = Gp + Ec*sparse(1:m,1:m,xhat)*Ec';
invG = G\I;

% compute inv_sqrtm(G)
invGsqrtm = sqrtm(G)\I;

% form the dual variable from a primal one
Y = invG*Qp*invG;

% dual variable initialization 
if gamma < 1.e-12
        
    % initialize the dual vector zphat
    zphat = 1.01*abs(gamma + qvec - sum(Ec.*(Y*Ec))') + 0.4*ones(m,1);
    zp = zphat;

    % initialize the dual vector zmhat
    zmhat = 1.01*abs(gamma - qvec + sum(Ec.*(Y*Ec))') + 0.2*ones(m,1);
    zm = zmhat;
    
else 
    
    % initialize the dual vector zphat
    zphat = 1.01*abs(gamma + qvec - sum(Ec.*(Y*Ec))');
    zp = zphat;

    % initialize the dual vector zmhat
    zmhat = 1.01*abs(gamma - qvec + sum(Ec.*(Y*Ec))');
    zm = zmhat;
    
end

% residual vector for the primal feasibility
rpvec = x - xp + xm;

% residual vectors for the dual feasibility
rd_p_vec = (gamma + qvec - sum(Ec.*(Y*Ec))') - zp; 
rd_m_vec = (gamma - qvec + sum(Ec.*(Y*Ec))') - zm; 

% Allocate memory for output variables
out.x   = zeros(m,1);  % optimal vector of edge weights (before polishing)
out.zp   = zeros(m,1); % optimal dual vector zp
out.zm   = zeros(m,1); % optimal dual vector zm
out.xp   = zeros(m,1); % optimal primal vector xp 
out.xm   = zeros(m,1); % optimal primal vector xm
out.dg  = zeros(length(opts.maxit),1); % duality gap
out.dresp = zeros(length(opts.maxit),1); % dual residual plus
out.dresm = zeros(length(opts.maxit),1); % dual residual minus
out.pres = zeros(length(opts.maxit),1); % dual residual
out.ind1 = zeros(length(opts.maxit),1); % number of PCG affine iterations
out.ind2 = zeros(length(opts.maxit),1); % number of PCG Mehrotra iterations

% duality gap
eta = xp'*zp + xm'*zm;

% parameter mu - quantifies the current duality gap
mu = eta/(2*m); 

% division between zp and xp
zp_over_xp = zphat./xphat;

% division between zm and xm
zm_over_xm = zmhat./xmhat;

% multiplication between zp and xp
vec_xpzp = xphat.*zphat;

% multiplication between zm and xm
vec_xmzm = xmhat.*zmhat;

% method type
if strcmp(flag,'chol'), % Cholesky factorization

    % matrix H
    H = 2*(Ec'*Y*Ec).*(Ec'*invG*Ec)';
       
    % matrix A
    % A = [ H + sparse(1:m,1:m,zp_over_xp), -H ; ...
    %      -H, H + sparse(1:m,1:m,zm_over_xm) ];
     
    % preconditioner   
    diagA = [diag(H) + zp_over_xp; diag(H) + zm_over_xm];
    % alternatively
    % diagA = [2*sum(Ec.*(Y*Ec))'.*sum(Ec.*(invG*Ec))' + zp_over_xp; ...
    %          2*sum(Ec.*(Y*Ec))'.*sum(Ec.*(invG*Ec))' + zm_over_xm];
    inv_diagA = 1./diagA;
    sqrt_inv_diagA = sqrt(inv_diagA);  
        
    % Cholesky factorization of A
    Alow = chol(sparse(1:2*m,1:2*m,sqrt_inv_diagA)* ...
               [ H + sparse(1:m,1:m,zp_over_xp), -H ; ...
                -H, H + sparse(1:m,1:m,zm_over_xm) ]* ...
                sparse(1:2*m,1:2*m,sqrt_inv_diagA),'lower');
           
    % indices ind1 and ind2
    ind1 = 1;
    ind2 = 1;
    
elseif strcmp(flag,'pcg1'), % PCG with storing A and diagonal preconditioner
    
    % matrix H
    H = 2*Ec'*Y*Ec.*(Ec'*invG*Ec)';
    
    % matrix A
    A = [ H + sparse(1:m,1:m,zp_over_xp), -H ; ...
         -H, H + sparse(1:m,1:m,zm_over_xm) ];
     
    % diagonal preconditioner
    diagA = [diag(H) + zp_over_xp; diag(H) + zm_over_xm];
    invdiagA = 1./diagA;
       
elseif strcmp(flag,'pcg2'), % PCG without storing A and diagonal preconditioner
   
    % diagonal preconditioner
    diagA = [2*sum(Ec.*(Y*Ec))'.*sum(Ec.*(invG*Ec))' + zp_over_xp; ...
             2*sum(Ec.*(Y*Ec))'.*sum(Ec.*(invG*Ec))' + zm_over_xm];
    invdiagA = 1./diagA;
    
else 
    
    error('Wrong method')
    
end

% counting index
s = 1;

% initiate variables
out.dg(s,:)  = eta;
out.pres(s,:) = norm(rpvec); % dual residual
out.dresp(s,:) = norm(rd_p_vec);
out.dresm(s,:) = norm(rd_m_vec);
out.x = xhat;
out.zp = zphat;
out.zm = zmhat;
out.xp = xphat;
out.xm = xmhat;
out.s = s;

% print
fprintf('%s %s %s %s %s %s %s\n\n', ' d_gap', '    p_res', '   d_res_p', '  d_res_m', ' i_1', 'i_2', '     iter');     

% Customized infeasible primal-dual IP algorithm
% based on Cholesky factorization
for s = 2:maxit,
       
    % method type
    if strcmp(flag,'chol'), % Cholesky factorization
        
        % compute affine scaling directions
        dxpm_a_tmp = (Alow')\(Alow\(sqrt_inv_diagA.* [-zphat - rd_p_vec + H*rpvec; ...
                      -zmhat - rd_m_vec - H*rpvec]));
                                
        dxpm_a = sqrt_inv_diagA.*dxpm_a_tmp;

        dxp_a = dxpm_a(1:m);
        dxm_a = dxpm_a(m+1:end);

        dzp_a = -(zp_over_xp.*dxp_a + zphat);
        dzm_a = -(zm_over_xm.*dxm_a + zmhat);
        
        % affine scaling directions for x
        dx_a = dxp_a - dxm_a - rpvec;
        
        % affine scaling directions for G
        % dG_a = Ec*sparse(1:m,1:m,dx_a)*Ec';
        
    elseif strcmp(flag,'pcg1'), % PCG with storing A and diagonal preconditioner
    
        % PCG Algorithm starts here
        % Algorithm 5.3 from Nocedal and Wright
        % same notation used
        
        % initial the affine search directions for PCG algorithm dxpm_a
        dxpm_a = zeros(2*m,1);
        
        % residual for affine scaling direction
        % rk = bvec - A*dxpm_a; % for dx_a \neq 0
        rk = ([-zphat - rd_p_vec + H*rpvec ; -zmhat - rd_m_vec - H*rpvec]); 
        
        % dk
        dk = rk.*invdiagA; 
        
        % A*dk 
        Adk = A*dk;
        
        % norm of b
        normb = norm([-zphat-rd_p_vec+H*rpvec; -zmhat-rd_m_vec-H*rpvec]);
  
        % affine scaling direction via PCG
        for ind1 = 1:maxit_pcg,
            
            % stopping criterion based on the relative duality gap
            if norm(rk) <= min(0.1,delta_pcg_aff*eta/normb),

                break

            end

            % compute q1 (an itermediate vector)
            q1 = (rk.*invdiagA);
            
            % compute alpha_k
            alpha_k = (rk'*q1)/(dk'*Adk);
            
            % update dxpm_a
            dxpm_a = dxpm_a + alpha_k*dk;
            
            % compute rknew
            rknew = rk - alpha_k*Adk;
            
            % compute q2 and q3 (intermediate vectors)
            q2 = q1 - alpha_k *(Adk.*invdiagA);
            q3 = ((rknew'*q2)/(rk'*q1));
            
            % update dk
            dk = q2 + q3*dk;
            
            % update A*dk
            Adk = A*q2 + q3*Adk;
            
            % update rk
            rk = rknew;

        end
        
        % update dxp_a and dxm_a
        dxp_a = dxpm_a(1:m);
        dxm_a = dxpm_a(m+1:end);
        
        % update dzp_a and dzm_a
        dzp_a = -(zp_over_xp.*dxp_a + zphat);
        dzm_a = -(zm_over_xm.*dxm_a + zmhat);
        
        % update dx_a and dG_a
        dx_a = dxp_a - dxm_a - rpvec;
        % dG_a = Ec*sparse(1:m,1:m,dx_a)*Ec';
        
    else % PCG without storing A and diagonal preconditioner
    
        % PCG Algorithm starts here
        % Algorithm 5.3 from Nocedal and Wright
        % same notation used
        
        % initial the affine search directions for PCG algorithm dxpm_a
        dxpm_a = zeros(2*m,1);
        
        % "middle" matrix Mmid1
        Mmid1 = Y*Ec*sparse(1:m,1:m,rpvec)*Ec'*invG;

        % residual for affine scaling direction
        % rk = bvec - A*dxpm_a; % for dx_a \neq 0
        rk = ([-zphat - rd_p_vec + 2*sum(Ec.*(Mmid1*Ec))'; ...
               -zmhat - rd_m_vec - 2*sum(Ec.*(Mmid1*Ec))']); 
        
        % dk
        dk = rk.*invdiagA; 
        
        % "middle" matrix Mmid11
        Mmid11 = Y*Ec*sparse(1:m,1:m,dk(1:m)-dk(m+1:end))*Ec'*invG;
        
        % A*dk
        Adk = [2*sum(Ec.*(Mmid11*Ec))' + zp_over_xp.*dk(1:m); ...
              -2*sum(Ec.*(Mmid11*Ec))' + zm_over_xm.*dk(m+1:end)];
               
        % norm of b
        normb = norm([-zphat - rd_p_vec + 2*sum(Ec.*(Mmid1*Ec))'; ...
                      -zmhat - rd_m_vec - 2*sum(Ec.*(Mmid1*Ec))' ]);
  
        % affine scaling direction via PCG
        for ind1 = 1:maxit_pcg,
           
            % stopping criterion based on the relative duality gap
            if norm(rk) <= min(0.1,delta_pcg_aff*eta/normb),

                break

            end

            % compute q1 (an itermediate vector)
            q1 = (rk.*invdiagA);
            
            % compute alpha_k
            alpha_k = (rk'*q1)/(dk'*Adk);
            
            % update dxpm_a
            dxpm_a = dxpm_a + alpha_k*dk;
            
            % compute rknew
            rknew = rk - alpha_k*Adk;
            
            % compute q2 and q3 (intermediate vectors)
            q2 = q1 - alpha_k *(Adk.*invdiagA);
            q3 = ((rknew'*q2)/(rk'*q1));
            
            % update dk
            dk = q2 + q3*dk;
            % update A*dk
            
            % "middle" matrix Mmid11
            Mmid11 = Y*Ec*sparse(1:m,1:m,dk(1:m)-dk(m+1:end))*Ec'*invG;
           
            % update Adk
            Adk = [2*sum(Ec.*(Mmid11*Ec))' + zp_over_xp.*dk(1:m); ...
                  -2*sum(Ec.*(Mmid11*Ec))' + zm_over_xm.*dk(m+1:end) ];
            
            % update rk
            rk = rknew;

        end
        
        % update dxp_a and dxm_a
        dxp_a = dxpm_a(1:m);
        dxm_a = dxpm_a(m+1:end);
        
        % update dzp_a and dzm_a
        dzp_a = -(zp_over_xp.*dxp_a + zphat);
        dzm_a = -(zm_over_xm.*dxm_a + zmhat);
        
        % update dx_a and dG_a
        dx_a = dxp_a - dxm_a - rpvec;
        % dG_a = Ec*sparse(1:m,1:m,dx_a)*Ec';
          
    end
      
    % step-size selection for affine search direction such that 
    % xp > 0, xm > 0, and G > 0 (see appendix B)
    
    % Ktilde = invGsqrtm*dG_a*invGsqrtm
    % Ktilde = invGsqrtm*Ec*sparse(1:m,1:m,dx_a)*Ec'*invGsqrtm;
     
    % eigen values of Ktilde
    [~,Dtilde] = eig(invGsqrtm*Ec*sparse(1:m,1:m,dx_a)*Ec'*invGsqrtm);
    % real part of Dtilde
    Etilde = real(diag(Dtilde));
    % minimum eigenvalue
    minEtilde = min(Etilde);
    
    if min(min(dxp_a./xphat),0) < 1.e-12,
        
        alpha_x1 = min(1,1/abs(min(dxp_a./(tau*xphat))));
        
    else
        
        alpha_x1 = 1;
        
    end
    
    if min(min(dxm_a./xmhat),0) < 1.e-12,
        
        alpha_x2 = min(1,1/abs(min(dxm_a./(tau*xmhat))));
        
    else
        
        alpha_x2 = 1;
        
    end
    
    if minEtilde < 1.e-12,
        
        alpha_x3 = min(1,1/abs(minEtilde));
        
    else
        
        alpha_x3 = 1;
        
    end
    
    alpha_x = min([alpha_x1,alpha_x2,alpha_x3]);
    
    
    if min(min(dzp_a./zphat),0) < 1.e-12,
        
        alpha_z1 = min(1,1/abs(min(dzp_a./(tau*zphat))));
        
    else
        
        alpha_z1 = 1;
        
    end
    
    if min(min(dzm_a./zmhat),0) < 1.e-12,
        
        alpha_z2 = min(1,1/abs(min(dzm_a./(tau*zmhat))));
        
    else
        
        alpha_z2 = 1;
        
    end
    
    alpha_z = min(alpha_z1,alpha_z2);
    
    % "centering" parameter sigma
    sigma = ( ( (xphat + alpha_x*dxp_a)'*(zphat + alpha_z*dzp_a) + ...
                (xmhat + alpha_x*dxm_a)'*(zmhat + alpha_z*dzm_a))/ ...
                (xphat'*zphat + xmhat'*zmhat) )^delta;
    
    % method type
    if strcmp(flag,'chol'), % Cholesky factorization

        % Mehrotra correction
        % (Alow')\(Alow\(sqrt_inv_diagA.*[(sigma*mu - ...
        % dxp_a.*dzp_a)./xphat;(sigma*mu - dxm_a.*dzm_a)./xmhat]));
        
        % search directions
        dxpm_tmp = dxpm_a_tmp + (Alow')\(Alow\(sqrt_inv_diagA.*[(sigma*mu - ...
            dxp_a.*dzp_a)./xphat;(sigma*mu - dxm_a.*dzm_a)./xmhat]));
        
        
        dxpm = sqrt_inv_diagA.*dxpm_tmp;
        
        dxp = dxpm(1:m);
        dxm = dxpm(m+1:end);
        
        % dx
        dx = dxp - dxm - rpvec;
        
        dzp = -zp_over_xp.*dxp + (sigma*mu - vec_xpzp - dxp_a.*dzp_a)./xphat;
        dzm = -zm_over_xm.*dxm + (sigma*mu - vec_xmzm - dxm_a.*dzm_a)./xmhat;
        
        % dG
        % dG = Ec*sparse(1:m,1:m,dx)*Ec';
        
    elseif strcmp(flag,'pcg1'), % PCG with storing A; diagonal preconditioner
        
        % initialization
        dxpm = dxpm_a;     
        
        % residual for scaling direction
        % rk = bvec_mc - A * dxpm; % for dx_mc \neq 0
        rk = ([(sigma*mu - dxp_a.*dzp_a)./xphat;(sigma*mu - dxm_a.*dzm_a)./xmhat]) - A*dxpm;
        
        % dk
        dk = rk.*invdiagA; 
        % A*dk
        Adk = A*dk;
        
        % norm of b
        normb = norm([(sigma*mu - dxp_a.*dzp_a)./xphat;(sigma*mu - dxm_a.*dzm_a)./xmhat]);
        
    % search direction via PCG
    for ind2 = 1:maxit_pcg,
           
            % stopping criterion based on the relative duality gap        
            if norm(rk) <= min(0.1,delta_pcg*eta/normb),

                break

            end

            % compute q1 (an intermediate vector)
            q1 = (rk.*invdiagA);
            
            % compute alpha_k
            alpha_k = (rk'*q1)/(dk'*Adk);
            
            % update dxpm
            dxpm = dxpm + alpha_k*dk;
            
            % update rknew
            rknew = rk - alpha_k*Adk;
            
            % compute intermediate vectors q2 and q3
            q2 = q1 - alpha_k *(Adk.*invdiagA);
            q3 = ((rknew'*q2)/(rk'*q1));
            
            % dk
            dk = q2 + q3*dk;
            
            % A*dk
            Adk = A*q2 + q3*Adk;
            
            %update rk
            rk = rknew;
        
    end
    
        % update dxpm
        dxpm = dxpm_a + dxpm;

        dxp = dxpm(1:m);
        dxm = dxpm(m+1:end);
        
        % update dx
        dx = dxp - dxm - rpvec;

        dzp = -zp_over_xp.*dxp + (sigma*mu - vec_xpzp - dxp_a.*dzp_a)./xphat;
        dzm = -zm_over_xm.*dxm + (sigma*mu - vec_xmzm - dxm_a.*dzm_a)./xmhat;
        
        % update dG
        % dG = Ec*sparse(1:m,1:m,dx)*Ec';
        
    else % PCG without storing A; diagonal preconditioner
        
        %initialization
        dxpm = dxpm_a;     
           
        % "middle" matrix Mmid1
        Mmid1 = Y*Ec*sparse(1:m,1:m,dxpm(1:m)-dxpm(m+1:end))*Ec'*invG;
        
        % A* dxpm
        Adxpm = [2*sum(Ec.*(Mmid1*Ec))' + zp_over_xp.*dxpm(1:m); ...
                -2*sum(Ec.*(Mmid1*Ec))' + zm_over_xm.*dxpm(m+1:end) ];
              
        % residual for scaling direction
        % rk = bvec_mc - A * dxpm; % for dx_mc \neq 0
        rk = ([(sigma*mu - dxp_a.*dzp_a)./xphat;(sigma*mu - dxm_a.*dzm_a)./xmhat]) - Adxpm;
        
        % dk
        dk = rk.*invdiagA;
        
        % A*dk
        % "middle" matrix Mmid11
        Mmid11 = Y*Ec*sparse(1:m,1:m,dk(1:m)-dk(m+1:end))*Ec'*invG;
        
        Adk = [2*sum(Ec.*(Mmid11*Ec))' + zp_over_xp.*dk(1:m); ...
              -2*sum(Ec.*(Mmid11*Ec))' + zm_over_xm.*dk(m+1:end)];
        
        % norm of b
        normb = norm([(sigma*mu - dxp_a.*dzp_a)./xphat;(sigma*mu - dxm_a.*dzm_a)./xmhat]);
        
    % search direction via PCG
    for ind2 = 1:maxit_pcg,
        
            % stopping criterion based on the relative duality gap        
            if norm(rk) <= min(0.1,delta_pcg*eta/normb),

                break

            end

            % compute q1 (an intermediate vector)
            q1 = (rk.*invdiagA);
            
            % compute alpha_k
            alpha_k = (rk'*q1)/(dk'*Adk);
            
            % update dxpm
            dxpm = dxpm + alpha_k*dk;
            
            % update rknew
            rknew = rk - alpha_k*Adk;
            
            % compute intermediate vectors q2 and q3
            q2 = q1 - alpha_k *(Adk.*invdiagA);
            q3 = ((rknew'*q2)/(rk'*q1));
            
            % dk
            dk = q2 + q3*dk;
                    
            % "middle" matrix Mmid11
            Mmid11 = Y*Ec*sparse(1:m,1:m,dk(1:m)-dk(m+1:end))*Ec'*invG;
            
            % A*dk
            Adk = [2*sum(Ec.*(Mmid11*Ec))' + zp_over_xp.*dk(1:m); ...
                  -2*sum(Ec.*(Mmid11*Ec))' + zm_over_xm.*dk(m+1:end)];
            
            % update rk
            rk = rknew;
        
    end
    
        % update dxpm
        dxpm = dxpm_a + dxpm;

        dxp = dxpm(1:m);
        dxm = dxpm(m+1:end);
        
        % update dx
        dx = dxp - dxm - rpvec;

        dzp = -zp_over_xp.*dxp + (sigma*mu - vec_xpzp - dxp_a.*dzp_a)./xphat;
        dzm = -zm_over_xm.*dxm + (sigma*mu - vec_xmzm - dxm_a.*dzm_a)./xmhat;
        
        % update dG
        % dG = Ec*sparse(1:m,1:m,dx)*Ec';
        
    end
            
    % step-size selection for affine search direction such that 
    % xp > 0, xm > 0, and G > 0 (see Appendix B)
    
    % Ktilde = invGsqrtm*dG*invGsqrtm;
    % Ktilde = invGsqrtm*Ec*sparse(1:m,1:m,dx)*Ec'*invGsqrtm;
   
    % eigenvalues of Ktilde
    [~,Dtilde] = eig(invGsqrtm*Ec*sparse(1:m,1:m,dx)*Ec'*invGsqrtm);
    % real part of Dtilde
    Etilde = real(diag(Dtilde));
    % minimum eigenvalue
    minEtilde = min(Etilde);
    
    if min(min(dxp./xphat),0) < 1.e-12,
        
        alpha_x1 = min(1,1/abs(min(dxp./(tau*xphat))));
        
    else
        
        alpha_x1 = 1;
        
    end
    
    if min(min(dxm./xmhat),0) < 1.e-12,
        
        alpha_x2 = min(1,1/abs(min(dxm./(tau*xmhat))));
        
    else
        
        alpha_x2 = 1;
        
    end
    
    if minEtilde < 1.e-12,
        
        alpha_x3 = min(1,1/abs(minEtilde));
        
    else
        
        alpha_x3 = 1;
        
    end
    
    alpha_x = min([alpha_x1,alpha_x2,alpha_x3]);
    
    if min(min(dzp./zphat),0) < 1.e-12,
        
        alpha_z1 = min(1,1/abs(min(dzp./(tau*zphat))));
        
    else
        
        alpha_z1 = 1;
        
    end
    
    if min(min(dzm./zmhat),0) < 1.e-12,
        
        alpha_z2 = min(1,1/abs(min(dzm./(tau*zmhat))));
        
    else
        
        alpha_z2 = 1;
        
    end
    
    alpha_z = min(alpha_z1,alpha_z2);
    
    % update xhat, xphat, xmhat and zphat, zmhat
    xphat = xphat + min(1,0.99*alpha_x)*dxp;
    xmhat = xmhat + min(1,0.99*alpha_x)*dxm;
    xhat  = xhat + min(1,0.99*alpha_x)*dx;
    zphat = zphat + min(1,0.99*alpha_z)*dzp;
    zmhat = zmhat + min(1,0.99*alpha_z)*dzm;
    
    % update x, xp, xm, zp, zm
    x  = xhat;
    xp = xphat;
    xm = xmhat;
    zp = zphat;
    zm = zmhat;
     
    % duality gap
    eta = (xp'*zp + xm'*zm);
    
    % G and invG
    G = Gp + Ec*sparse(1:m,1:m,xhat)*Ec';
    invG = G\I;
    
    % compute inv_sqrtm(G)
    invGsqrtm = sqrtm(G)\I;
    
    % form the dual variable from a primal one
    Y = invG*Qp*invG;
    
    % residual vector for the primal feasibility
    rpvec = x - xp + xm;

    % residual vectors for the dual feasibility
    rd_p_vec = gamma + qvec - sum(Ec.*(Y*Ec))' - zp; 
    rd_m_vec = gamma - qvec + sum(Ec.*(Y*Ec))' - zm;  
    
    % variable used in the stopping criterion
    check = any( [eta > eps_dg, norm(rpvec) > eps_pres, norm(rd_p_vec) > eps_dres, norm(rd_m_vec) > eps_dres] );
    
    % declare variables
    out.dg(s,:)  = eta;
    out.pres(s,:) = norm(rpvec);
    out.dresp(s,:) = norm(rd_p_vec);
    out.dresm(s,:) = norm(rd_m_vec);
    out.ind1(s-1) = ind1;
    out.ind2(s-1) = ind2;
    out.x = xhat;
    out.zp = zphat;
    out.zm = zmhat;
    out.xp = xphat;
    out.xm = xmhat;
    out.s = s;
  
    % display
    disp([num2str(eta,'%6.1E'),'   ', ...
        num2str(norm(rpvec),'%6.1E'),'   ', ...
        num2str(norm(rd_p_vec),'%6.1E'),'   ', ...
        num2str(norm(rd_m_vec),'%6.1E'),'   ', ...
        num2str(ind1,'%i'),'   ', ...
        num2str(ind2,'%i'),'        ' ...
        num2str(s-1,'%i')])
        
    if check == 0
        
        break;
        
    else
        
        % parameter mu - quantifies the current duality gap
        mu = eta/(2*m);
        
        % division between zp and xp
        zp_over_xp = zp./xp;

        % division between zm and xm
        zm_over_xm = zm./xm;
        
        %multiplication of zp and xp
        vec_xpzp = xp.*zp;
        
        %multiplication of zm and xm
        vec_xmzm = xm.*zm;
        
        if strcmp(flag,'chol'),
             
            % matrix H
            H = 2*Ec'*Y*Ec.*(Ec'*invG*Ec)';

            % matrix A
            % A = [ H + sparse(1:m,1:m,zp_over_xp), -H ; ...
            %      -H, H + sparse(1:m,1:m,zm_over_xm) ];
             
            % preconditioner   
            diagA = [diag(H) + zp_over_xp; diag(H) + zm_over_xm];
            % alternatively
            % diagA = [2*sum(Ec.*(Y*Ec))'.*sum(Ec.*(invG*Ec))' + zp_over_xp; ...
            %          2*sum(Ec.*(Y*Ec))'.*sum(Ec.*(invG*Ec))' + zm_over_xm];
            inv_diagA = 1./diagA;
            sqrt_inv_diagA = sqrt(inv_diagA);
            
            % Cholesky factorization of A
            Alow = chol(sparse(1:2*m,1:2*m,sqrt_inv_diagA)* ...
                [ H + sparse(1:m,1:m,zp_over_xp), -H ; ...
                -H, H + sparse(1:m,1:m,zm_over_xm) ]* ...
                sparse(1:2*m,1:2*m,sqrt_inv_diagA),'lower');
                    
        elseif strcmp(flag,'pcg1'),
             
            % matrix H
            H = 2*Ec'*Y*Ec.*(Ec'*invG*Ec)';
            
            % matrix A
            A = [ H + sparse(1:m,1:m,zp_over_xp), -H ; ...
                 -H, H + sparse(1:m,1:m,zm_over_xm) ];
             
            % diagonal preconditioner
            diagA = [diag(H) + zp_over_xp; diag(H) + zm_over_xm];
            invdiagA = 1./diagA;
            
        else

            % diagonal preconditioner
            diagA = [2*sum(Ec.*(Y*Ec))'.*sum(Ec.*(invG*Ec))' + zp_over_xp; 2*sum(Ec.*(Y*Ec))'.*sum(Ec.*(invG*Ec))' + zm_over_xm];
            invdiagA = 1./diagA;
          
        end

    end  
    
end

