%%
% Matlab script that explains how to run graphsp_IP
%
% Graph type: flag_p = 1; % Erdos-Renyi network
%             flag_p = 2; % path network
%             flag_p = 3; % disconnected randomly generated network (with N = 50 nodes)
%
% Outputs:
%
% out1 (for opts.method = 'chol'); t1 (total running time);
% out2 (for opts.method = 'pcg1'); t2 (total running time);
% out3 (for opts.method = 'pcg2'); t3 (total running time);
% out4 (for proxBB); t4 (total running time);
% out5 (for quadCD); t5 (total running time).
%
clear, clc, close all

% number of nodes
N = 50;

flag_p = 1; % for Erdos-Renyi network
% flag_p = 2; % for path network
% flag_p = 3; % for disconnected randomly generated network (with N = 50 nodes)

if flag_p == 1,
    
    % Edge probability for Erdos-Renyi network
    p = 1.05*log(N)/N;
    
    % Graph Laplacian of an Erdos-Renyi plant network
    Lp = random_graph_generator(N,p);
    eigLp = sort(eig(Lp),'ascend');

%     % Check if a plant graph is connected 
%     while abs(eigLp(2)) <= 1.e-12
% 
%         Lp = random_graph_generator(N,p);
%         eigLp = sort(eig(Lp),'ascend');
% 
%         if abs(eigLp(2)) >= 1.e-12
%             
%             break;
%             
%         end
%         
%     end
     
elseif flag_p == 2

    % Construct incidence matrix of a path graph
    Ep = toeplitz([1 -1 zeros(1,N-2)],[1 zeros(1,N-2)]);
    
    % plant graph Laplacian
    Lp = Ep*Ep';
      
elseif flag_p == 3,
    
    N = 50;
    Length = 10;

    load pos;
    pos = pos(1:N,:);
    r = Length/5;
    reweighted_Max_Iter = 1;

    m = N*(N-1)/2;

    % determine the neighbors of each node
    idx = zeros(2,m);
    Lp = zeros(N,N); 
    stp = 0;

    % construct Lp
    for i = 1:N,

        for j = i+1:N,
            % if the distance between two nodes is less than r
            % then they are neighbors.

            if norm( pos(i,:) - pos(j,:), 2 ) <= r,
                
                stp = stp + 1;
                idx(:,stp) = [i j]';
                Lp(i,j) = -1;
                Lp(j,i) = -1;
                
            end

        end

    end

    for p = 1:N

       Lp(p,p) = -sum(Lp(p,:));

    end
    
else 
    
     error('Incorrect graph type')
     
end

% make sure Lp is sparse
Lp = sparse(Lp);

% Construct incidence matrix of the controller graph
% No overlap between plan and controller
% Ecomplete = [Ep, Ec]
d = diag(Lp);
S = N^2 - N + nnz(d) - nnz(Lp);
index = zeros(S/2,2);

c = 1;

for i = 1:N,
    
    for j = i+1:N,
        
        if Lp(i,j)==0,
            
            index(c,:) = [i,j];
            
            c = c + 1;
            
        end
        
    end
    
end

% incidence matrix of the controller graph
Ec = incmat(index',N);

% number of edges in the controller graph
m = size(Ec,2);
    
% state and control weights in LQR objective
% state weight - deviation from average
Q = eye(N) - (1/N)*ones(N,N);
% control weight
R = speye(N);

% "strengthened" state weight
Qp = Q + (1/N)*ones(N,N) + Lp*R*Lp; 

% choose gamma
if (abs(eigLp(2)) > 1.e-6 || flag_p == 2) 
    
    % compute Gp
    % Gp = Lp + (1/N)*ones(N,N);
    invGp = (Lp + (1/N)*ones(N,N))\eye(N);
    Yp = invGp*Q*invGp;
    
    % compute gamma_max
    gamma_max = max(abs(sum(Ec.*(Yp*Ec))));
    % regularization parameter gamma
    gamma = 0.8*full(gamma_max);

else
    
    % for random disconnected network
    gamma = 1000;
    
end


% run graphsp_IP

% options for Cholesky factorization
opts = struct('eps_dg',1.e-3,'eps_pres',1.e-6,'maxit',100,'maxit_pcg',2000,'method','chol','delta_pcg_aff',0.3,'delta_pcg',0.3);

% code based on Cholesky factorization
tic, 
    out1 = graphsp_IP(Lp,Ec,Q,R,gamma,opts);
t1 = toc;

% options for PCG method with storing matrix A
opts = struct('eps_dg',1.e-3,'eps_pres',1.e-6,'maxit',100,'maxit_pcg',2000,'method','pcg1','delta_pcg_aff',0.3,'delta_pcg',0.3);

% code based on PCG method
% matrix A stored in memory; diagonal preconditioner
tic, 
    out2 = graphsp_IP(Lp,Ec,Q,R,gamma,opts);
t2 = toc;

% options for PCG method without storing matrix A
opts = struct('eps_dg',1.e-3,'eps_pres',1.e-6,'maxit',100,'maxit_pcg',2000,'method','pcg2','delta_pcg_aff',0.3,'delta_pcg',0.3);

% code based on PCG method
% matrix A NOT stored in memory; diagonal preconditioner
tic, 
    out3 = graphsp_IP(Lp,Ec,Q,R,gamma,opts);
t3 = toc;

% options for sequential quadratic approximation method
opts = struct('maxit',1000,'eps_pres',1.e-6,'eps_dres',1.e-1,'eps_dg',1.e-3,'flagBB',1);

% code based on proxN
tic,

out4 = graphsp_proxBB(Lp,Ec,R,gamma,opts);
t4 = toc;

% options for sequential quadratic approximation method
opts = struct('maxit',200,'eps_pres',1.e-6,'eps_dres',1.e-1,'eps_dg',1.e-3);

% code based on proxN
tic,

out5 = graphsp_proxN(Lp,Ec,R,gamma,opts);
t5 = toc;



% running times
% chol vs pcg1 vs pcg2 vs proxBB vs proxN
[t1 t2 t3 t4 t5]

  