% Sparsity-Promoting Optimal H_2 Controller for Undirected Networks
%
% Written by: Sepideh Hassan-Moghaddam and Mihailo R. Jovanovic
%
% May 2015
%
% Customized Algorithm based on Infeasible Primal-Dual Interior-Point Method
%
% Network Dynamics: qdot = -(Lp + Lx)*q + d
% 
% q - state of the network 
% d - white stochastic disturbance 
%
% Lp - graph Laplacian of the plant
% Lx = Ec*diag(x)*Ec' - graph Laplacian of the controller
% Ec - incidence matrix of the controller graph
% x  - vector of the edge weights
%
% Optimization Problem
%
% minimize <Qp,(Gp + Lx)^{-1}> + <R,Lx> + gamma*sum(w_i x_i)
%
% subject to Gp + Lx > 0
% 
% Syntax:
%
% out = out = graphsp_IP_w(Lp,Ec,Q,R,gamma,W,x0,xp0,xm0,opts) 
%
% Inputs: (1) plant graph Laplacian Lp
%             incidence matrix of the controller graph Ec
%             state weight Q
%             controller weight R
%             sparsity-promoting parameter gamma
%             weight matrix W
%             starting point for x:  x0
%             starting point for xp: xp0
%             starting point for xm: xm0
%
%         (2) options
%
%             opts.eps_dg   - tolerance on duality gap
%             opts.eps_dres - tolerance on dual residual
%             opts.eps_pres - tolerance on primal residual
%             opts.maxit    - maximum number of IP iterations
%
%             Search Directions Obtained Using: 
%
%             Cholesky Factorization
%
%             If opts argument is omitted, the default values are set to
%
%             opts.eps_dg   = 1.e-6
%             opts.eps_dres = 1.e-3
%             opts.eps_pres = 1.e-6
%             opts.maxit    = 100
%
% Output:  out - output structure containing
%
%          out.x    - optimal vector of edge weights 
%          out.zp   - optimal z+
%          out.zm   - optimal z-
%          out.xp   - optimal x+
%          out.xm   - optimal x-
%          out.dg   - duality gap
%          out.dres - dual residual
%          out.pres - primal residual
%          out.s    - number of IP iterations
%          out.ind1 - number of PCG iterations for affine search direction
%          out.ind2 - number of PCG iterations for Mehrotra search direction
%
% Additional information:
%
% http://www.umn.edu/~mihailo/software/graphsp/


function out = graphsp_IP_w(Lp,Ec,Q,R,gamma,W,x0,xp0,xm0,opts) 

% Initialization
if nargin < 9
    error('At least six input arguments are required.')
elseif nargin == 9
    opts = struct('eps_dg',1.e-6,'eps_dres',1.e-3,'eps_pres',1.e-6,'maxit',100);
elseif nargin > 10
    error('Too many input arguments.')
end

% number of nodes
N = size(Ec,1);

% number of edges in the controller graph
% determines the number of optimization variables
m = size(Ec,2);

% Data preprocessing
eps_dg  = opts.eps_dg;  % tolerance on duality gap
eps_dres = opts.eps_dres; % tolerance on dual residual
eps_pres = opts.eps_pres; % tolerance on primal residual
maxit = opts.maxit;

% sparse identity matrix
I = speye(N); 

% delta in IP method
delta = 3;

% tau for step size selection
if gamma <= 1.e-12
    
    % step-size parameter for optimal centralized controller
    tau = 0.9;
    
else
    
    % step-size parameter for sparsity-promoting problem
    tau = 1;
    
end
 
% "strengthened" plant graph Laplacian
Gp = Lp + (1/N)*ones(N,N);

% "strengthened" state weight 
Qp = Q + (1/N)*ones(N,N) + Lp*R*Lp; 

% form the vector that appears in the objective function
qvec = sum(Ec.*(R*Ec))';

% initialize feedback gain
xhat = x0; 
x = xhat;
 
% initialize xp
xphat = xp0; 
xp = xphat;

% initialize xm
xmhat = xm0; 
xm = xmhat;

% compute invG
G = Gp + Ec*sparse(1:m,1:m,xhat)*Ec';
invG = G\I;

% compute inv_sqrtm(G)
invGsqrtm = sqrtm(G)\I;

% form the dual variable from a primal one
Y = invG*Qp*invG;

% dual variable initialization 
if gamma < 1.e-12
 
    % initialize the dual vector zphat
    zphat = 1.01*abs(gamma*W + qvec - sum(Ec.*(Y*Ec))') + 0.4*ones(m,1);
    zp = zphat;

    % initialize the dual vector zmhat
    zmhat = 1.01*abs(gamma*W - qvec + sum(Ec.*(Y*Ec))') + 0.2*ones(m,1);
    zm = zmhat;
    
else 
    
    % initialize the dual vector zphat
    zphat = 1.01*abs(gamma*W + qvec - sum(Ec.*(Y*Ec))');
    zp = zphat;

    % initialize the dual vector zmhat
    zmhat = 1.01*abs(gamma*W - qvec + sum(Ec.*(Y*Ec))');
    zm = zmhat;
    
end

% residual vector for the primal feasibility
rpvec = x - xp + xm;

% residual vectors for the dual feasibility
rd_p_vec = (gamma*W + qvec - sum(Ec.*(Y*Ec))') - zp; 
rd_m_vec = (gamma*W - qvec + sum(Ec.*(Y*Ec))') - zm; 

% Allocate memory for output variables
out.x   = zeros(m,1); % optimal vector of edge weights (before polishing)
out.zp   = zeros(m,1); % optimal dual vector zp
out.zm   = zeros(m,1); % optimal dual vector zm
out.xp   = zeros(m,1); % optimal primal vector xp 
out.xm   = zeros(m,1); % optimal primal vector xm
out.dg  = zeros(length(opts.maxit),1); % duality gap
out.dresp = zeros(length(opts.maxit),1); % dual residual plus
out.dresm = zeros(length(opts.maxit),1); % dual residual minus
out.pres = zeros(length(opts.maxit),1); % dual residual
out.ind1 = zeros(length(opts.maxit),1); % number of PCG affine iterations
out.ind2 = zeros(length(opts.maxit),1); % number of PCG Mehrotra iterations

% duality gap
eta = xp'*zp + xm'*zm;

% parameter mu - quantifies the current duality gap
mu = eta/(2*m); 

% division between zp and xp
zp_over_xp = zphat./xphat;

% division between zm and xm
zm_over_xm = zmhat./xmhat;

% multiplication between zp and xp
vec_xpzp = xphat.*zphat;

% multiplication between zm and xm
vec_xmzm = xmhat.*zmhat;
   
% matrix H
H = 2*(Ec'*Y*Ec).*(Ec'*invG*Ec)';

% matrix A
% A = [ H + sparse(1:m,1:m,zp_over_xp), -H ; ...
%      -H, H + sparse(1:m,1:m,zm_over_xm) ];

% preconditioner
diagA = [diag(H) + zp_over_xp; diag(H) + zm_over_xm];
% alternatively
% diagA = [2*sum(Ec.*(Y*Ec))'.*sum(Ec.*(invG*Ec))' + zp_over_xp; ...
%          2*sum(Ec.*(Y*Ec))'.*sum(Ec.*(invG*Ec))' + zm_over_xm];
inv_diagA = 1./diagA;
sqrt_inv_diagA = sqrt(inv_diagA);

% Cholesky factorization of A
Alow = chol(sparse(1:2*m,1:2*m,sqrt_inv_diagA)* ...
    [ H + sparse(1:m,1:m,zp_over_xp), -H ; ...
    -H, H + sparse(1:m,1:m,zm_over_xm) ]* ...
    sparse(1:2*m,1:2*m,sqrt_inv_diagA),'lower');
    
% indices ind1 and ind2
ind1 = 1;
ind2 = 1;
    
% counting index
s = 1;

% initiate variables
out.dg(s,:)  = eta;
out.pres(s,:) = norm(rpvec); % dual residual
out.dresp(s,:) = norm(rd_p_vec);
out.dresm(s,:) = norm(rd_m_vec);
out.x = xhat;
out.zp = zphat;
out.zm = zmhat;
out.xp = xphat;
out.xm = xmhat;
out.s = s;

% display
fprintf('%s %s %s %s %s %s %s\n\n', ' d_gap', '    p_res', '   d_res_p', '  d_res_m', ' i_1', 'i_2', '     iter');     

% Customized infeasible primal-dual IP algorithm
% based on Cholesky factorization
for s = 2:maxit,
        
    % compute affine scaling directions
    dxpm_a_tmp = (Alow')\(Alow\(sqrt_inv_diagA.* [-zphat - rd_p_vec + H*rpvec; ...
                  -zmhat - rd_m_vec - H*rpvec]));
                                
    dxpm_a = sqrt_inv_diagA.* dxpm_a_tmp;

    dxp_a = dxpm_a(1:m);
    dxm_a = dxpm_a(m+1:end);

    dzp_a = -(zp_over_xp.*dxp_a + zphat);
    dzm_a = -(zm_over_xm.*dxm_a + zmhat);
        
    % affine scaling directions for x
    dx_a = dxp_a - dxm_a - rpvec;
        
    % affine scaling directions for G
    % dG_a = Ec*sparse(1:m,1:m,dx_a)*Ec';
         
    % step-size selection for affine search direction such that 
    % xp > 0, xm > 0, and G > 0 (see appendix B)
   
    % eigenvalues of Ktilde
    [~,Dtilde] = eig(invGsqrtm*Ec*sparse(1:m,1:m,dx_a)*Ec'*invGsqrtm);
    % real part of Dtilde
    Etilde = real(diag(Dtilde));
    % minimum eigenvalue
    minEtilde = min(Etilde);
    
    if min(min(dxp_a./xphat),0) < 1.e-12,
        
        alpha_x1 = min(1,1/abs(min(dxp_a./(tau*xphat))));
        
    else
        
        alpha_x1 = 1;
        
    end
    
    if min(min(dxm_a./xmhat),0) < 1.e-12,
        
        alpha_x2 = min(1,1/abs(min(dxm_a./(tau*xmhat))));
        
    else
        
        alpha_x2 = 1;
        
    end
    
    if minEtilde < 1.e-12,
        
        alpha_x3 = min(1,1/abs(minEtilde));
        
    else
        
        alpha_x3 = 1;
        
    end
    
    alpha_x = min([alpha_x1,alpha_x2,alpha_x3]);
    
    
    if min(min(dzp_a./zphat),0) < 1.e-12,
        
        alpha_z1 = min(1,1/abs(min(dzp_a./(tau*zphat))));
        
    else
        
        alpha_z1 = 1;
        
    end
    
    if min(min(dzm_a./zmhat),0) < 1.e-12,
        
        alpha_z2 = min(1,1/abs(min(dzm_a./(tau*zmhat))));
        
    else
        
        alpha_z2 = 1;
        
    end
    
    alpha_z = min(alpha_z1,alpha_z2);
    
    % "centering" parameter sigma
    sigma = ( ( (xphat + alpha_x*dxp_a)'*(zphat + alpha_z*dzp_a) + ...
                (xmhat + alpha_x*dxm_a)'*(zmhat + alpha_z*dzm_a))/ ...
                (xphat'*zphat + xmhat'*zmhat) )^delta;
    
    % Mehrotra correction
    % (Alow')\(Alow\(sqrt_inv_diagA.*[(sigma*mu - ...
    % dxp_a.*dzp_a)./xphat;(sigma*mu - dxm_a.*dzm_a)./xmhat]));
         
    % search directions
    dxpm_tmp = dxpm_a_tmp + (Alow')\(Alow\(sqrt_inv_diagA.*[(sigma*mu - ...
               dxp_a.*dzp_a)./xphat;(sigma*mu - dxm_a.*dzm_a)./xmhat]));


    dxpm = sqrt_inv_diagA.*dxpm_tmp;

    dxp = dxpm(1:m);
    dxm = dxpm(m+1:end);
    
    % dx
    dx = dxp - dxm - rpvec;
    
    dzp = -zp_over_xp.*dxp + (sigma*mu - vec_xpzp - dxp_a.*dzp_a)./xphat;
    dzm = -zm_over_xm.*dxm + (sigma*mu - vec_xmzm - dxm_a.*dzm_a)./xmhat;
    
    % dG
    % dG = Ec*sparse(1:m,1:m,dx)*Ec';
        
    % step-size selection for search direction such that 
    % xp > 0, xm > 0, and G > 0 (see appendix B)
    
    % Ktilde
    % Ktilde = invGsqrtm*Ec*sparse(1:m,1:m,dx)*Ec'*invGsqrtm;
    
    % eigen values of Ktilde
    [~,Dtilde] = eig(invGsqrtm*Ec*sparse(1:m,1:m,dx)*Ec'*invGsqrtm);
    % real part of Dtilde
    Etilde = real(diag(Dtilde));
    % minimum eigenvalue
    minEtilde = min(Etilde);
    
    if min(min(dxp./xphat),0) < 1.e-12,
        
        alpha_x1 = min(1,1/abs(min(dxp./(tau*xphat))));
        
    else
        
        alpha_x1 = 1;
        
    end
    
    if min(min(dxm./xmhat),0) < 1.e-12,
        
        alpha_x2 = min(1,1/abs(min(dxm./(tau*xmhat))));
        
    else
        
        alpha_x2 = 1;
        
    end
    
    if minEtilde < 1.e-12,
        
        alpha_x3 = min(1,1/abs(minEtilde));
        
    else
        
        alpha_x3 = 1;
        
    end
    
    alpha_x = min([alpha_x1,alpha_x2,alpha_x3]);
    
    if min(min(dzp./zphat),0) < 1.e-12,
        
        alpha_z1 = min(1,1/abs(min(dzp./(tau*zphat))));
        
    else
        
        alpha_z1 = 1;
        
    end
    
    if min(min(dzm./zmhat),0) < 1.e-12,
        
        alpha_z2 = min(1,1/abs(min(dzm./(tau*zmhat))));
        
    else
        
        alpha_z2 = 1;
        
    end
    
    alpha_z = min(alpha_z1,alpha_z2);
    
    % update xhat, xphat, xmhat and zphat, zmhat
    xphat = xphat + min(1,0.99*alpha_x)*dxp;
    xmhat = xmhat + min(1,0.99*alpha_x)*dxm;
    xhat  = xhat + min(1,0.99*alpha_x)*dx;
    zphat = zphat + min(1,0.99*alpha_z)*dzp;
    zmhat = zmhat + min(1,0.99*alpha_z)*dzm;
    
    % update x, xp, xm, zp, zm
    x = xhat;
    xp = xphat;
    xm = xmhat;
    zp = zphat;
    zm = zmhat;
     
    % duality gap
    eta = xp'*zp + xm'*zm;
    
    % G and invG
    G = Gp + Ec*sparse(1:m,1:m,xhat)*Ec';
    invG = G\I;
    
    % compute inv_sqrtm(G)
    invGsqrtm = sqrtm(G)\I;
    
    % form the dual variable from a primal one
    Y = invG*Qp*invG;
    
    % residual vector for the primal feasibility
    rpvec = x - xp + xm;

    % residual vectors for the dual feasibility
    rd_p_vec = gamma*W + qvec - sum(Ec.*(Y*Ec))' - zp; 
    rd_m_vec = gamma*W - qvec + sum(Ec.*(Y*Ec))' - zm;  
    
    % variable used in the stopping criterion
    check = any( [eta > eps_dg, norm(rpvec) > eps_pres, norm(rd_p_vec) > eps_dres, norm(rd_m_vec) > eps_dres] );
  
    % declare variables
    out.dg(s,:)  = eta;
    out.pres(s,:) = norm(rpvec);
    out.dresp(s,:) = norm(rd_p_vec);
    out.dresm(s,:) = norm(rd_m_vec);
    out.ind1(s-1) = ind1;
    out.ind2(s-1) = ind2;
    out.x = xhat;
    out.zp = zphat;
    out.zm = zmhat;
    out.xp = xphat;
    out.xm = xmhat;
    out.s = s;
  
    % display
    disp([num2str(eta,'%6.1E'),'   ', ...
        num2str(norm(rpvec),'%6.1E'),'   ', ...
        num2str(norm(rd_p_vec),'%6.1E'),'   ', ...
        num2str(norm(rd_m_vec),'%6.1E'),'   ', ...
        num2str(ind1,'%i'),'   ', ...
        num2str(ind2,'%i'),'        ' ...
        num2str(s-1,'%i')])
        
    if check == 0
        
        break;
        
    else
        
        % parameter mu - quantifies the current duality gap
        mu = eta/(2*m);
        
        % division between zp and xp
        zp_over_xp = zp./xp;

        % division between zm and xm
        zm_over_xm = zm./xm;
        
        %multiplication of zp and xp
        vec_xpzp = xp.*zp;
        
        %multiplication of zm and xm
        vec_xmzm = xm.*zm;
        
        % Cholesky factorization of the matrix A
        % matrix H
        H = 2*(Ec'*Y*Ec).*(Ec'*invG*Ec)';
        
        % matrix A
        % A = [ H + sparse(1:m,1:m,zp_over_xp), -H ; ...
        %      -H, H + sparse(1:m,1:m,zm_over_xm) ];
        
        % preconditioner
        diagA = [diag(H) + zp_over_xp; diag(H) + zm_over_xm];
        % alternatively
        % diagA = [2*sum(Ec.*(Y*Ec))'.*sum(Ec.*(invG*Ec))' + zp_over_xp; ...
        %          2*sum(Ec.*(Y*Ec))'.*sum(Ec.*(invG*Ec))' + zm_over_xm];
        inv_diagA = 1./diagA;
        sqrt_inv_diagA = sqrt(inv_diagA);
        
        % Cholesky factorization of A
        Alow = chol(sparse(1:2*m,1:2*m,sqrt_inv_diagA)* ...
            [ H + sparse(1:m,1:m,zp_over_xp), -H ; ...
            -H, H + sparse(1:m,1:m,zm_over_xm) ]* ...
            sparse(1:2*m,1:2*m,sqrt_inv_diagA),'lower');
          
    end  
    
end

