
% Sparsity-Promoting Optimal H_2 Controller for Growing Connected Resistive Networks
%
% Written by: Sepideh Hassan-Moghaddam and Mihailo R. Jovanovic
%
% February 2016
%
% Proximal gradient with positivity constraints 
% (i.e., no soft-thresholding)

% Network Dynamics: qdot = -(Lp + Lx)*q + d
% 
% q - state of the network 
% d - white stochastic disturbance 
%
% Lp - graph Laplacian of the plant
% Lx = Ec*diag(x)*Ec' - graph Laplacian of the controller
% Ec - incidence matrix of the controller graph
% x  - vector of the edge weights
%
% Optimization Problem
%
% minimize <Qp,(Gp + Lx)^{-1}> + <R,Lx> + gamma*sum(x_i)
% 
% subject to x_i >= 0
% 
% Syntax:
%
% out = graphsp_proxBB_res(Lp,Ec,Q,R,gamma,opts)
%
% Inputs: (1) plant graph Laplacian Lp
%             incidence matrix of the controller graph Ec
%             state weight Q
%             controller weight R
%             sparsity-promoting parameter gamma
%
%         (2) options
%
%             opts.eps_dg    - tolerance on duality gap
%             opts.eps_res   - tolerance on dual residual
%             opts.maxit     - maximum number of IP iterations
%             opts.flagBB    - flag for BB step-size initialization

%
%             If opts argument is omitted, the default values are set to
%
%             opts.eps_dg    = 1.e-4
%             opts.eps_res   = 1.e-3
%             opts.maxit     = 1000
%             opts.flagBB    = 1
%
% Output:  out - output structure containing
%
%          out.x    - optimal vector of edge weights 
%          out.s    - number of proxBB iterations
%          out.dg   - duality gap
%          out.res  - dual residual
%          out.ss   - step-size
%          out.js   - number of iterations for line-search
%          out.iter_time - time of an inner iteration
%
% Additional information:
%
% http://www.umn.edu/~mihailo/software/graphsp/

function out = graphsp_proxBB_res(Lp,Ec,R,gamma,opts) 

% Initialization
if nargin < 4
    error('At least four input arguments are required.')
elseif nargin == 4
    opts = struct('maxit',1000,'eps_res',1.e-3,'eps_dg',1.e-4,'flagBB',1);
elseif nargin > 5
    error('Too many input arguments.')
end

% number of nodes
N = size(Ec,1);

% number of edges in the controller graph
% determines the number of optimization variables
m = size(Ec,2);

% Initialization
I = eye(N);

% Data preprocessing
maxit = opts.maxit;
eps_dg = opts.eps_dg;
eps_res = opts.eps_res;
flagBB = opts.flagBB;

% number of line search iterations (1 means no backtracking)
ls = 1;

% backtracking parameters
ssize = 1;

% "strengthened" plant graph Laplacian
Gp = Lp + (1/N)*ones(N,N);

% strengthened state weight 
% Qp = Q + (1/N)*ones(N,N) + Lp*R*Lp; 
% for Q = Qda = I - (1/N)*ones(N,N)
Qp = I + Lp*R*Lp; 

% % for Q = I - (1/N)*ones(N,N)
% Qp = I + Lp*R*Lp; 

% form the vector that appears in the objective function
qvec = sum(Ec.*(R*Ec))';

% initialize feedback gain
x = 0.5*ones(m,1); 
% x = rand(m,1);

% compute invG
% G = Gp + Ec*sparse(1:m,1:m,x)*Ec';
invG = (Gp + Ec*sparse(1:m,1:m,x)*Ec')\I;

% form dual variable from a primal one
Y = invG*Qp*invG;

% grad of J
gradJ = qvec + gamma - sum(Ec.*(Y*Ec))';

% Allocate memory for gamma-dependent output variables
out.x = zeros(m,length(gamma)); % vector of amplitudes (before polishing)
out.s = zeros(length(gamma)); % number of proxBB iterations
out.dg  = zeros(length(gamma),length(opts.maxit)); % duality gap
out.res = zeros(length(gamma),length(opts.maxit)); % dual residual
out.ss  = zeros(length(gamma),length(opts.maxit)); % step-size in each iteration
out.js  = zeros(length(gamma),length(opts.maxit)); % number of iterations for line-search

% display
fprintf('%s %s %s  \n\n', 's','   d_gap', '    d_res');  

%  Customized proximal gradient algorithm
for s = 1:maxit,
    
    % save time of the iteration
    start1 = tic;
    
    % line search
    
    % change this ls to the number of iterations you need for line search
    for j = 1:ls,
        
        % indicator function with positivity constraints
        xnew = max(x - ssize*gradJ,0);
        
        % invGnew
        invGnew = (Gp + Ec*sparse(1:m,1:m,xnew)*Ec')\I;
        
%         % objective function
%         Jnew = trace(invGnew*Qp) + (qvec + gamma)'*xnew;     
% 
%       % backtracking line search (if needed)
%       if Jnew <= J - ssize*gradJ'*GradJ + ...
%                  0.5*ssize*(norm(GradJ)^2),
% 
%           break;
% 
%       else
% 
%           ssize = ssize*beta;
% 
%       end

        
    end

    
    % record stepsize
    out.ss(s) = ssize;
   
    % form a dual variable 
    Ynew = invGnew*Qp*invGnew;
    
    % grad of J
    gradJnew = qvec + gamma - sum(Ec.*(Ynew*Ec))';
    
    % use BB stepsize initialization
    if flagBB == 1,
        
        % BB stepsize initialization
        ssize = (norm(xnew - x)^2) / ...
            ((x - xnew)'*(gradJ - gradJnew));
        
        if ssize < 0 || ssize == Inf,
            
            ssize = 1;
            
        end
        
    else
        
        ssize = 1;
        
    end
    
    % change of variable to guarantee dual feasibility
    temp = min(gradJnew);
    
    if (temp >= 0)
    
        % dual feasible z
        z = gradJnew;
        
    else
        
        % beta
        beta1 = (gamma + 2*R(1,1))/(max(-gradJnew + gamma)+ 2*R(1,1));
        
        % dual feasible Y (change of variable)
        Yprime = beta1 * Ynew + ((1-beta1)/N)*ones(N,N);
        
        % dual feasible z
        z = qvec + gamma - sum(Ec.*(Yprime*Ec))';
        
    end
    
    % duality gap
    eta = xnew'*z;

    % dual residual
    res_vec = gradJnew - z; 
    
    % stopping criteria
    check = any( [eta > eps_dg, norm(res_vec) > eps_res]);
    
    % declare variables
    out.dg(1,s)  = eta;
    out.res(1,s) = norm(res_vec);
    out.js(s) = j;

    
    % check convergence
    if check == 0
       
        % update x
        x = xnew;

    % display duality gap, dual residual
           disp([num2str(s),'   ', ...
             num2str(eta,'%6.1E'),'   ', ...
             num2str(norm(res_vec),'%6.1E'),'   ', ...
             ])
        
       % time of an inner iteration
        out.iter_time(s) = toc(start1);
        
        break;
        
    end
    
    % update x 
    x = xnew;
    
    % update gradJ
    gradJ = gradJnew;
    
    % time of an inner iteration
    out.iter_time(s) = toc(start1);
    
    % display duality gap, dual residual
         disp([num2str(s),'   ', ...
             num2str(eta,'%6.1E'),'   ', ...
             num2str(norm(res_vec),'%6.1E'),'   ', ...
             ])
        
    
end

% Record output data
out.s = s;
out.x = x; % vector of amplitudes

