% Sparsity-Promoting Optimal H_2 Controller for Growing Connected Resistive Networks
%
% Written by: Sepideh Hassan-Moghaddam and Mihailo R. Jovanovic
%
% February 2015
%
% Coordinate Descent Algorithm based on 
% Sequential Quadratic Approximation
% Active set strategy used
%
% Network Dynamics: qdot = -(Lp + Lx)*q + d
% 
% q - state of the network 
% d - white stochastic disturbance 
%
% Lp - graph Laplacian of the plant
% Lx = Ec*diag(x)*Ec' - graph Laplacian of the controller
% Ec - incidence matrix of the controller graph
% x  - vector of the edge weights
%
% Optimization Problem
%
% minimize <Qp,(Gp + Lx)^{-1}> + <R,Lx> + gamma*sum(x_i)
% 
% subject to x_i >= 0
% 
% Syntax:
%
% out = graphsp_proxN_res(Lp,Ec,Q,R,gamma,opts)
%
% Inputs: (1) plant graph Laplacian Lp
%             incidence matrix of the controller graph Ec
%             state weight Q
%             controller weight R
%             sparsity-promoting parameter gamma
%
%         (2) options
%
%             opts.eps_dg    - tolerance on duality gap
%             opts.eps_res   - tolerance on dual residual
%             opts.maxit     - maximum number of proxN iterations
%
%             If opts argument is omitted, the default values are set to
%
%             opts.eps_dg    = 1.e-4
%             opts.eps_res   = 1.e-3
%             opts.maxit     = 200
%
% Output:  out - output structure containing
%
%          out.x    - optimal vector of edge weights 
%          out.s    - number of proxN iterations
%          out.dg   - duality gap
%          out.res  - dual residual
%          out.js   - number of iterations for line-search
%          out.ss   - step-size
%          out.iter_time - time of an inner iteration
%
% Additional information:
%
% http://www.umn.edu/~mihailo/software/graphsp/

function out = graphsp_proxN_res(Lp,Ec,R,gamma,opts)

% Initialization
if nargin < 4
    error('At least four input arguments are required.')
elseif nargin == 4
    opts = struct('maxit',200,'eps_dg',1.e-4,'eps_res',1.e-3);
elseif nargin > 5
    error('Too many input arguments.')
end

% number of nodes
N = size(Ec,1);

% number of edges in the controller graph
% determines the number of optimization variables
m = size(Ec,2);

% Initialization
I = speye(N);

% Data preprocessing
maxit = opts.maxit;
eps_dg = opts.eps_dg;
eps_res= opts.eps_res;

% backtracking parameters
alpha = 0.3;
beta  = 0.5;
tau = 1;

tol3 = 0.001;

% "strengthened" plant graph Laplacian
Gp = Lp + (1/N)*ones(N,N);

% strengthened state weight
% Qp = Q + (1/N)*ones(N,N) + Lp*R*Lp;
% for Q = Qda = I - (1/N)*ones(N,N)
Qp = I + Lp*R*Lp;

% % for Q = I - (1/N)*ones(N,N)
% Qp = I + Lp*R*Lp;

% form the vector that appears in the objective function
qvec = sum(Ec.*(R*Ec))';

% initialize feedback gain
x = zeros(m,1);
% x = rand(m,1);

% invG
% G = Gp + Ec * diag(x) * Ec';
invG = (Gp + Ec*sparse(1:m,1:m,x)*Ec')\I;

% form dual variable from a primal one
Y = invG*Qp*invG;

% gradient of J
Jgrad = sum(Ec.*(Y*Ec))';

% diagonal of a Hessian
diagH = Jgrad.*sum(Ec.*(invG*Ec))';
Jgrad = qvec - Jgrad;

% gradient of f
fgrad = Jgrad + gamma;

% quadratic approximation of J
% phi = 0.5*x'*H*x + fgrad'*x;

% objective function
J = trace(invG*Qp) + qvec'*x;
f = J + gamma*sum(x);

% all indices 
ind_all = 1:m;

% Allocate memory for gamma-dependent output variables
out.x = zeros(m,length(gamma)); % vector of amplitudes (before polishing)
out.s = zeros(length(gamma)); % number of proxN iterations
out.dg  = zeros(length(gamma),length(opts.maxit)); % duality gap
out.res  = zeros(length(gamma),length(opts.maxit)); % dual residual
out.ss  = zeros(length(gamma),length(opts.maxit)); % step-size
out.js  = zeros(length(gamma),length(opts.maxit)); % Armijo iterations

% initialize eta
eta = 100;

% display
fprintf('%s %s %s  \n\n', 's','   d_gap', '    d_res');  

for s = 1:maxit,
    
    % save time of the iteration
    start1 = tic;
    
    % search direction
    dx = zeros(m,1);
    
    % norm of dx (using in stopping criterion)
    normdx = 0;
    
    % sum of mu's
    musum = 0;
    
    % H*dx
    Hdx = zeros(m,1);

    % active set strategy   
    postindex = find(fgrad >= 0);
    
    % find zero elements in x
    zerox = find(x == 0);
    fixedindex = intersect(zerox, postindex);
    
    % active set
    index = setdiff(ind_all,fixedindex);
    
    
    % going through the active set and update the directions
    for ind = 1:2*m,
        
        % going through the active set
        for iin = 1:size(index,2),
            
            % ith element of the active set
            i = index(iin);
            
            % b/diagH(i)
            b_over_diagH = (Hdx(i) + fgrad(i))/diagH(i);
            
            % c
            c = x(i) + dx(i);
            cba = c - b_over_diagH;
            
            % find the update mu for the search direction dx
            if ( cba >= 0)
                
                mu = -b_over_diagH;
                
            else
                
                mu = -c;
                
            end
            
            % update dx
            dx(i) = dx(i) + mu;
            
            % norm of dx
            normdx = normdx + abs(dx(i));
            
            % sum of mu's
            musum = musum + abs(mu);
            
            % Hessian times dx
            Hdx = Hdx + mu*(Ec'*(Y*Ec(:,i))).*((invG*Ec(:,i))'*Ec)';
             
        end
       
        
        % stopping criterion of the inner iteration
        % measures how big are the changes in each direction compared to norm of the vector of edge weights.
        if ( musum <= normdx * tol3)
            
            break;
            
        end
        
        % reset
        musum = 0;
        normdx = 0;
        
    end

    
    % positivity of x + Armijo rule
    for j = 1:1000,
        
        % update x
        xnew = x + tau*dx;
        
        % update G
        Gnew = Gp + Ec*sparse(1:m,1:m,xnew)*Ec';
        
        % check positivity
        if min(xnew)<0
            
            tau = tau*beta;
            
        else
            
            % invGnew
            invGnew = Gnew\I;
            
            % form dual variable from a primal one
            Ynew = invGnew*Qp*invGnew;
            
            % objective function
            Jnew = trace(invGnew*Qp) + qvec'*xnew;
            fnew = Jnew + gamma*sum(xnew);
            
            % Armijo rule
            if fnew <= f + tau*alpha*( fgrad'*dx );
                
                break;
                
            end
            
            tau = tau*beta;
            
        end
        
    end
    
    % declare variables
    out.js(s) = j;
    out.ss(s) = tau;

    % update f
    f = fnew;

    % update x
    x = xnew;

    % update invG
    invG = invGnew;

    % form dual variable from a primal one
    Y = Ynew;

    % gradient of J
    Jgrad = sum(Ec.*(Y*Ec))';
    % diagonal of a Hessian
    diagH = Jgrad.*sum(Ec.*(invG*Ec))';
    Jgrad = qvec - Jgrad;
    % gradient of f
    fgrad = Jgrad + gamma;

    % reset tau
    tau = 1;

    % time of an inner iteration
    out.iter_time(s) = toc(start1);
    
    
    % change of variable to guarantee dual feasibility    
    temp = min(fgrad);
    
    if (temp >= 0)
        
        % dual feasible z
        z = fgrad;
        
    else
        
        % beta
        beta1 = 0.999*(gamma + 2*R(1,1))/(max(-Jgrad)+ 2*R(1,1));
        
        % dual feasible Y (change of variable)
        Yprime = beta1 * Y + ((1-beta1)/N)*ones(N,N);
        
        % dual feasible z
        z = qvec + gamma - sum(Ec.*(Yprime*Ec))';
        
    end
    
    % duality gap
    etanew = x'*z;

    % residual vector for the dual feasibility
    res_vec = fgrad - z; 
    
    
    % stopping criterion
    check = any( [etanew > eps_dg, norm(res_vec) > eps_res]);
    % declare variables
    out.dg(s)  = etanew;
    out.res(s) = norm(res_vec);

    if check == 0
        
        
       % display duality gap, dual residual
           disp([num2str(s),'   ', ...
             num2str(etanew,'%6.1E'),'   ', ...
             num2str(norm(res_vec),'%6.1E'),'   ', ...
             ])
        
        
        break;
    end
    
    if abs(eta-etanew)<1.e-8
        
      % display duality gap, dual residual
           disp([num2str(s),'   ', ...
             num2str(etanew,'%6.1E'),'   ', ...
             num2str(norm(res_vec),'%6.1E'),'   ', ...
             ])
         
        break;
    end
    
    eta = etanew;
    
    
    
      % display duality gap, dual residual
           disp([num2str(s),'   ', ...
             num2str(etanew,'%6.1E'),'   ', ...
             num2str(norm(res_vec),'%6.1E'),'   ', ...
             ])
    

    
end
out.x = x; % vector of amplitudes
out.s = s; % number of iterations

