%%
% Matlab script that explains how to run graphsp_IP_res
%
% Graph type: flag_p = 1; % for Erdos-Renyi network
%             flag_p = 2; % for path network
%
% Regularization parameter set to 0.5*gamma_max
%
% Outputs:
%
% out1 (for opts.method = 'chol?); t1 (total running time);
% out2 (for opts.method = 'pcg1?); t2 (total running time);
% out3 (for opts.method = 'pcg2?); t3 (total running time).
%

clear, clc, close all

% number of nodes
N = 50;

% Network type
flag_p = 1; % for Erdos-Renyi network
% flag_p = 2; % for path network

if flag_p == 1
    
    % Edge probability for Erdos-Renyi network
    p = 1.05*log(N)/N;
    
    % Graph Laplacian of an Erdos-Renyi plant network
    Lp = random_graph_generator(N,p);
    eigLp = sort(eig(Lp),'ascend');

    % Check if a plant graph is connected 
    while abs(eigLp(2)) <= 1.e-12

        Lp = random_graph_generator(N,p);
        eigLp = sort(eig(Lp),'ascend');

        if abs(eigLp(2)) >= 1.e-12
            
            break;
            
        end
        
    end
 
elseif flag_p == 2,

    % Construct incidence matrix of a path graph
    Ep = sparse(toeplitz([1 -1 zeros(1,N-2)],[1 zeros(1,N-2)]));
    
    % Plant graph Laplacian
    Lp = Ep*Ep';
    
else 
    
     error('Incorrect graph type')
     
end

% Construct incidence matrix of the controller graph
% No overlap between plan and controller
% Ecomplete = [Ep, Ec]
% feel free to experiment with other choices for Ec
d = diag(Lp);
S = N^2 - N + nnz(d) - nnz(Lp);
index = zeros(S/2,2);

c = 1;

for i = 1:N,
    
    for j = i+1:N,
        
        if Lp(i,j)==0,
            
            index(c,:) = [i,j];
            
            c = c + 1;
            
        end
        
    end
    
end

% incidence matrix of the controller graph
Ec = incmat(index',N);

% number of edges in the controller graph
m = size(Ec,2);
 
% state and control weights in LQR objective
% state weight - deviation from average
Q = eye(N) - (1/N)*ones(N,N);
% control weight
R = speye(N);

% "strengthened" state weight
Qp = Q + (1/N)*ones(N,N) + Lp*R*Lp; 
   
% compute Gp
% Gp = Lp + (1/N)*ones(N,N);
invGp = (Lp + (1/N)*ones(N,N))\eye(N);
Yp = invGp*Q*invGp;

% compute gamma_max
gamma_max = max(abs(sum(Ec.*(Yp*Ec))));
% regularization parameter gamma
gamma = 0.5*full(gamma_max);

% run graphsp_IP_res

% options for IP method based on Cholesky factorization
opts = struct('eps_dg',1.e-6,'eps_res',1.e-3,'maxit',100,'maxit_pcg',1000,'method','chol','delta_pcg_aff',0.3,'delta_pcg',0.3);

% code based on Cholesky factorization
tic, 
    out1 = graphsp_IP_res(Lp,Ec,Q,R,gamma,opts);
t1 = toc;

% options for IP method based on PCG method pcg1
opts = struct('eps_dg',1.e-6,'eps_res',1.e-3,'maxit',100,'maxit_pcg',1000,'method','pcg1','delta_pcg_aff',.3,'delta_pcg',.3);

% code based on pcg1 
% matrix A stored in memory; diagonal preconditioner
tic, 
    out2 = graphsp_IP_res(Lp,Ec,Q,R,gamma,opts);
t2 = toc;

% options for IP method based on PCG method pcg2
opts = struct('eps_dg',1.e-6,'eps_res',1.e-3,'maxit',100,'maxit_pcg',1000,'method','pcg2','delta_pcg_aff',.3,'delta_pcg',.3);

% code based on pcg2 
% matrix A NOT stored in memory; diagonal preconditioner
tic, 
    out3 = graphsp_IP_res(Lp,Ec,Q,R,gamma,opts);
t3 = toc;

% running times
% chol vs pcg1 vs pcg2
[t1 t2 t3]

% Visualize results
figure
% plot optimal vector of the edge weights (before polishing)
plot(1:m,out1.x,'*',1:m,out2.x,'o',1:m,out3.x,'+')
xlab = xlabel('edge index','interpreter','tex');
set(xlab,'FontName','cmmr10','FontSize',26)
ylab = ylabel('edge weights','interpreter','tex');
set(ylab,'FontName','cmmr10','FontSize',26)
h = get(gcf,'CurrentAxes');
set(h,'FontName','cmr10','FontSize',20,'xscale','lin','yscale','lin')
axis([0 m+1 0 1.1*max(max([out1.x out2.x out3.x]))])

figure
% plot duality gaps
semilogy(1:out1.s,out1.dg,'*-',1:out2.s,out2.dg,'o-',1:out3.s,out3.dg,'*-')
xlab = xlabel('iteration index','interpreter','tex');
set(xlab,'FontName','cmmr10','FontSize',26)
ylab = ylabel('duality gap','interpreter','tex');
set(ylab,'FontName','cmmr10','FontSize',26)
h = get(gcf,'CurrentAxes');
set(h,'FontName','cmr10','FontSize',20,'xscale','lin','yscale','log')
axis([1 max([out1.s out2.s out2.s]) ...
      0.1*min([min(out1.dg) min(out2.dg) min(out3.dg)]) ...
      10*max([max(out1.dg) max(out2.dg) max(out3.dg)])])

figure
% plot primal residuals
semilogy(1:out1.s,out1.res,'*-',1:out2.s,out2.res,'o-',1:out3.s,out3.res,'*-')
xlab = xlabel('iteration index','interpreter','tex');
set(xlab,'FontName','cmmr10','FontSize',26)
ylab = ylabel('primal residual','interpreter','tex');
set(ylab,'FontName','cmmr10','FontSize',26)
h = get(gcf,'CurrentAxes');
set(h,'FontName','cmr10','FontSize',20,'xscale','lin','yscale','log')
axis([1 max([out1.s out2.s out2.s]) ...
      0.1*min([min(out1.res) min(out2.res) min(out3.res)]) ...
      10*max([max(out1.res) max(out2.res) max(out3.res)])])

% set controller edge weight to zero 
% (if they are below given threshold)
ind = find(out3.x <= 1.e-4);
x = out3.x;
x(ind) = 0;

figure
% plot structure of plant and controller graphs
% blue circles - plan graph
% red stars - controller graph
spy(Lp,'bo')
hold
spy(Ec*sparse(1:m,1:m,x)*Ec','r*')
hold
