%%
% Matlab script that explains how to run graphsp_proxN_res and graphsp_proxBB_res
%
% Graph type: flag_p = 1; % for Erdos-Renyi network
%             flag_p = 2; % for path network
%
% Regularization parameter set to 0.8*gamma_max
%
% Outputs:
%
% out1 (for proxN); t1 (total running time);
% out2 (for proxBB); t2 (total running time).
%

clear, clc, close all

% number of nodes
N = 100;

% Network type
flag_p = 1; % for Erdos-Renyi network
% flag_p = 2; % for path network

if flag_p == 1
    
    % Edge probability for Erdos-Renyi network
    p = 1.05*log(N)/N;
    
    % Graph Laplacian of an Erdos-Renyi plant network
    Lp = random_graph_generator(N,p);
    eigLp = sort(eig(Lp),'ascend');

    % Check if a plant graph is connected 
    while abs(eigLp(2)) <= 1.e-6

        Lp = random_graph_generator(N,p);
        eigLp = sort(eig(Lp),'ascend');

        if abs(eigLp(2)) >= 1.e-6
            
            break;
            
        end
        
    end
 
elseif flag_p == 2,

    % Construct incidence matrix of a path graph
    Ep = sparse(toeplitz([1 -1 zeros(1,N-2)],[1 zeros(1,N-2)]));
    
    % Plant graph Laplacian
    Lp = Ep*Ep';
    
else 
    
     error('Incorrect graph type')
     
end

% Construct incidence matrix of the controller graph
% No overlap between plan and controller
% Ecomplete = [Ep, Ec]
% feel free to experiment with other choices for Ec
d = diag(Lp);
S = N^2 - N + nnz(d) - nnz(Lp);
index = zeros(S/2,2);

c = 1;

for i = 1:N,
    
    for j = i+1:N,
        
        if Lp(i,j)==0,
            
            index(c,:) = [i,j];
            
            c = c + 1;
            
        end
        
    end
    
end

% incidence matrix of the controller graph
Ec = incmat(index',N);

% number of edges in the controller graph
m = size(Ec,2);
 
% state and control weights in LQR objective
% state weight - deviation from average
Q = eye(N) - (1/N)*ones(N,N);
% control weight
R = speye(N);

% "strengthened" state weight
Qp = Q + (1/N)*ones(N,N) + Lp*R*Lp; 


% regularization parameter gamma
if N < 1200,
    
    % compute invGp and Yp
    invGp = (Lp + (1/N)*ones(N,N))\eye(N);
    Yp = invGp*Q*invGp;
    
    % compute gamma_max
    gamma_max = max(abs(sum(Ec.*(Yp*Ec))));
    gamma = 0.8*gamma_max;

else
    
    gamma = 1.8;
    
end

gamma = full(gamma);

% run graphsp_proxN_res

% % options for proxN method
 opts = struct('maxit',200,'eps_dg',1.e-4,'eps_res',1.e-3);
% 
%  code based on proxN
tic, 
     out1 = graphsp_proxN_res(Lp,Ec,R,gamma,opts);
t1 = toc;

% run graphsp_proxBB_res

% % options for proxBB method
 opts = struct('maxit',1000,'eps_res',1.e-3,'eps_dg',1.e-4,'flagBB',1);
% 
% % code based on proxBB
tic, 
    out2 = graphsp_proxBB_res(Lp,Ec,R,gamma,opts);
t2 = toc;

% running times
% proxN vs proxBB
[t1 t2]

% Visualize results
figure
% plot optimal vector of the edge weights (before polishing)
plot(1:m,out1.x,'*',1:m,out2.x,'o')
xlab = xlabel('edge index','interpreter','tex');
set(xlab,'FontName','cmmr10','FontSize',26)
ylab = ylabel('edge weights','interpreter','tex');
set(ylab,'FontName','cmmr10','FontSize',26)
h = get(gcf,'CurrentAxes');
set(h,'FontName','cmr10','FontSize',20,'xscale','lin','yscale','lin')
axis([0 m+1 0 1.1*max(max([out1.x out2.x]))])

figure
% plot duality gaps
semilogy(1:out1.s,out1.dg,'*-',1:out2.s,out2.dg,'o-')
xlab = xlabel('iteration index','interpreter','tex');
set(xlab,'FontName','cmmr10','FontSize',26)
ylab = ylabel('duality gap','interpreter','tex');
set(ylab,'FontName','cmmr10','FontSize',26)
h = get(gcf,'CurrentAxes');
set(h,'FontName','cmr10','FontSize',20,'xscale','lin','yscale','log')
axis([1 max([out1.s out2.s]) ...
      0.1*min([min(out1.dg) min(out2.dg)]) ...
      10*max([max(out1.dg) max(out2.dg)])])

figure
% plot primal residuals
semilogy(1:out1.s,out1.res,'*-',1:out2.s,out2.res,'o-')
xlab = xlabel('iteration index','interpreter','tex');
set(xlab,'FontName','cmmr10','FontSize',26)
ylab = ylabel('primal residual','interpreter','tex');
set(ylab,'FontName','cmmr10','FontSize',26)
h = get(gcf,'CurrentAxes');
set(h,'FontName','cmr10','FontSize',20,'xscale','lin','yscale','log')
axis([1 max([out1.s out2.s]) ...
      0.1*min([min(out1.res) min(out2.res)]) ...
      10*max([max(out1.res) max(out2.res)])])

% set controller edge weight to zero 
% (if they are below given threshold)
ind = find(out2.x <= 1.e-4);
x = out2.x;
x(ind) = 0;

figure
% plot structure of plant and controller graphs
% blue circles - plan graph
% red stars - controller graph
spy(Lp,'bo')
hold
spy(Ec*sparse(1:m,1:m,x)*Ec','r*')
hold








